
/* Portions Copyright (c) 1995,1996,1997 by Trusted Information Systems, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND TRUSTED INFORMATION SYSTEMS DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL TRUSTED INFORMATION 
 * SYSTEMS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 *
 * Trusted Information Systems, Inc. has received approval from the
 * United States Government for export and reexport of TIS/DNSSEC
 * software from the United States of America under the provisions of
 * the Export Administration Regulations (EAR) General Software Note
 * (GSN) license exception for mass market software.  Under the
 * provisions of this license, this software may be exported or
 * reexported to all destinations except for the embargoed countries of
 * Cuba, Iran, Iraq, Libya, North Korea, Sudan and Syria.  Any export
 * or reexport of TIS/DNSSEC software to the embargoed countries
 * requires additional, specific licensing approval from the United
 * States Government. 
 */




static char *rcsid = "$Header: /home/ogud/sec_bind/dnssec_lib/RCS/sign_lib.c,v 1.32 1997/01/22 21:33:33 ogud Exp $";

/* 
 * Sign_lib.c: Routines used in the signing and verifiation process
 * Author: Olafur Gudmundsson ogud@tis.com 
 */

#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/param.h>
#include <string.h>
#include <memory.h>
#include <netinet/in.h> 
#include "../include/arpa/nameser.h"
#include "named.h"
#include "dnssec.h"
#ifdef DNS_SECURITY
#include "../include/resolv.h" 

static int         do_verify        __P(( char *, short, char *, 
			        	    struct databuf *, KEY *, 
				            u_int32_t));
static KEY        *get_pub_key      __P(( char *, u_int16_t, int *));
static int         set_sig_prefix   __P(( struct databuf *));
static void        verify_del       __P(( struct namebuf *, char *, 
					  int16_t type, struct databuf *));

/* 
 * nt_lookup( name, type, np) 
 * looks for the first record of name and type and returns pointer to it. 
 * If no record is found we will return NULL 
 * This routine does not work properly if ROUND_ROBIN is turned on 
 * ROUND_ROBIN does not work with security any way and should be outlawed
 * from secure servers 
 * Lookup for T_SIG will fail 
 */ 

struct databuf * 
nt_lookup( name, type, nnp)
const char *name;
short type; 
struct namebuf **nnp;
{
    struct namebuf  *np;
    struct databuf  *dp;
    const char      *fname;
    struct hashbuf  *htp=hashtab;          /* lookup in authorative data */
 

    if( type == T_AXFR || type == T_SIG)
      return NULL ;  /* we can not find which SIG rr is requested and
			there is no T_AXFR record only SIG(AXFR) */
    *nnp = np = nlookup( name, &htp, &fname, 0);
    if (np == NULL || np->n_data == NULL) {  /* name not found look in cache */
      htp = fcachetab;
      *nnp = np = nlookup ( name, &htp, &fname, 0);
      if ( np == NULL || np->n_data == NULL) 
	return NULL; 
    }

    if( type == 0)
      return np->n_data;
#ifdef ROUND_ROBIN
    for(dp = np->n_data; dp && dp->d_type != type ; dp=dp->d_next); 
    return dp; 
#else 
    for(dp = np->n_data; dp && (dp->d_type < type || dp->d_type == T_SIG) ;)
	dp=dp->d_next; 
    if( dp && dp->d_type == type)
      return dp; 
   else return NULL;
#endif 
}

static struct databuf * dp_rrarray[1024];

void
append_rr( domain, dp, ttl)
     char * domain;
     struct databuf     *dp;
     u_int32_t ttl;
{
    u_int32_t  t_ttl, opt;
    int i;
    t_ttl = dp->d_ttl;
    dp->d_ttl = ttl;
    dnptrs[0]= NULL;   /*  do not compress names */
    opt = _res.options;
    _res.options |=  RES_TRANS_LOWER;
    i = make_rr(domain, dp, &sig_data.data[sig_data.dat_len],
		sizeof(sig_data.data)-sig_data.dat_len, 0);
    _res.options = opt;
    dp->d_ttl = t_ttl;
    if( i <= 0){
      dprintf(10,(ddt, "append_rr: make_rr returned error code %s %s %d\n",
		  domain, (char *) __p_type(dp->d_type), i));
      return;
    }
    else {
      dp_rrarray[dp_rrcount++] = dp;
      sig_data.dat_len += i;
      sig_data.data[sig_data.dat_len] = 0x0;
    }
}

void 
expand_rrs( domain)
char *domain;
{
  u_int32_t t_ttl, opt;
  int i, len;
 /* array is sorted before we get here */ 
  for( i = 0; i < dp_rrcount; i++) { 
    t_ttl = dp_rrarray[i]->d_ttl;
    dp_rrarray[i]->d_ttl = sig_data.ttl;
    dnptrs[0]= NULL;   /*  do not compress names */
    opt = _res.options;
    _res.options |=  RES_TRANS_LOWER;
    len = make_rr(domain, dp_rrarray[i], &sig_data.data[sig_data.dat_len],
		sizeof(sig_data.data)-sig_data.dat_len, 0);
    _res.options = opt;
    dp_rrarray[i]->d_ttl = t_ttl;
    if( len <= 0){
      dprintf(10,(ddt, "expand_rr: make_rr returned error code %s %s %d\n",
		  domain, (char *) __p_type(dp_rrarray[i]->d_type), len));
      return;
    }
    else {
      sig_data.dat_len += len;
      sig_data.data[sig_data.dat_len] = 0x0;
    }
  }
}
/*
 * put_rr(): routine to put the data record into an data area for sign/verify
 * function to work with
 * domain = domain name of the record if domain is NULL this is
 *	    subsequent RR and no need to set the prefix
 * dp     = pointer to the RR we are to sign
 *  sign_ttl = the TTL to use for verification 
 * This routine will not walk through the databuf cain.
 */
void
put_rr(domain, dp, sign_ttl)
     char           *domain;
     struct databuf     *dp;
     u_int32_t sign_ttl;
{
  int j,n;
  sig_data.ttl = sign_ttl;
  for( j = dp_rrcount ; j > 0  && (n=db_cmp(dp, dp_rrarray[j-1])) < 0; j--) 
    dp_rrarray[j] = dp_rrarray[j-1];
  dp_rrarray[j] = dp;
  dp_rrcount++;
}

/*
 * set_prefix(); this routine creates the signature record prefix 
 */


/* 
 * set_sig_prefix(): This function takes the data from a signature record 
 * and puts it in the sig_data.prefix variable 
 * Returns size of the prefix.
 */
static int
set_sig_prefix( dp)
struct databuf *dp;
{
  int len;
  u_long opt;
  memcpy(sig_data.prefix, dp->d_data, SIGN_SIGNER); /* got all fields but 
						      the signers name */
  opt = _res.options;
  _res.options |=  RES_TRANS_LOWER;
  len = dn_comp( &dp->d_data[SIGN_SIGNER], &sig_data.prefix[SIGN_SIGNER],
		   sizeof(sig_data.prefix)-SIGN_SIGNER, NULL, NULL);
  _res.options = opt;
  sig_data.pre_len = SIGN_SIGNER + len ;
/* I only set prefix when I have new data delete old data*/
  if(key_file==0) 
    sig_data.dat_len = 0; 
  return sig_data.pre_len;
}


/* 
 * do_verify(): function to perform the actual verification of records
 *      that have been put in sig_data structure with a given signature
 *      record.
 *      Is aware of AXFR processing. 
 */
static int 
do_verify(domain, type, signer, sdp, key, sign_ttl)
     char *domain, *signer;
     KEY *key;
     short type;
     struct databuf *sdp;
     u_int32_t sign_ttl;
{
  int len, ret = 0; 
  u_char *cp = NULL; 

  if( type == T_SIG ) 
     return NOT_OK;    /* SIG(SIG) is not allowed */ 
  if( dp_rrcount == 0 && sig_data.dat_len == 0 && (type != T_AXFR)) {
    memset(sig_data.data, 0, sig_data.dat_len+8);
    sig_data.dat_len = dp_rrcount = 0; 
    return NOT_OK;  
  }
  
  cp =  sdp->d_data + SIGN_SIGNER + strlen(&sdp->d_data[SIGN_SIGNER])+1;
  len = sdp->d_size + sdp->d_data- cp;
  __putshort(type, sig_data.prefix); /*set the type we are doing VERIFY over */
  if( type == T_AXFR && key->axfr) { 
    if( key_file) {
      if((ret = dnssec_verify((SIG_MODE_AXFR|SIG_MODE_INIT), sig_data.prefix, 
			    sig_data.pre_len, key, NULL, 0)))
	dprintf(4,(ddt,"do_verify:Verify Failed for keyfile INIT %s %s %s %d\n"
		    , domain, signer, __p_type(type), ret)); 
      if((ret = dnssec_verify((SIG_MODE_AXFR|SIG_MODE_FINAL), sig_data.data, 
			    sig_data.dat_len, key, cp, len)))
	dprintf(4,(ddt,"do_verify:Verify Failed for keyfile FINAL %s %s %s %d\n"
		    , domain, signer, __p_type(type), ret)); 
      else {
	dprintf(33, (ddt, "do_verify:Verify Success %s %s %s %d\n", 
		     domain, signer,__p_type(type), ret)); 
#ifdef TESTING
	printf("Do_verify:Verify Success %s %s %s %d\n", domain, signer,
	     __p_type(type), ret); 
#endif /* TESTING */
      }
    }
    else {
      if((ret = dnssec_verify((SIG_MODE_AXFR | SIG_MODE_FINAL),
			 NULL, 0, key, cp, len)))
	dprintf(4, (ddt, "do_verify:Verify Failed %s %s %s %d\n", domain, 
		    signer,  __p_type(type), ret)); 
      else {
	dprintf(33, (ddt, "do_verify:Verify Success %s %s %s %d\n", 
		     domain, signer,__p_type(type), ret)); 
#ifdef TESTING
	printf("Do_verify:Verify Success %s %s %s %d\n", domain, signer,
	     __p_type(type), ret); 
#endif /* TESTING */
      }
    }
  }
  else {  /* if not T_AXFR  check if we want to do SIG(AXFR) */
    if( do_axfr &&( type == T_SOA || 
	  (key_file && key->axfr))){
      __putshort(T_AXFR, sig_data.prefix);
      if((ret = dnssec_verify((SIG_MODE_AXFR|SIG_MODE_INIT), 
			   sig_data.prefix, sig_data.pre_len, key, NULL, 0)))
	dprintf(4, (ddt, "do_verify:Verify Failed %s %s %s %d\n", domain, signer, 
	     __p_type(type), ret)); 
      key->axfr = do_axfr;
      __putshort(type, sig_data.prefix);
    }
    
    expand_rrs(domain);
    /* Normal signature processing */ 
    if((ret = dnssec_verify( SIG_MODE_INIT, sig_data.prefix, sig_data.pre_len,
				  key, cp, len)) ||
       (ret = dnssec_verify( SIG_MODE_FINAL, sig_data.data, sig_data.dat_len,
				  key, cp, len)))
      dprintf(4, (ddt, "do_verify:Verify Failed %s %s %s %d\n", domain, 
		  signer, __p_type(type), ret)); 
    else { 
/*
 * now check if the signature has expired if signature fails I can not use 
 *  it to delete data (denial of service attack) 
 */
      u_int32_t exp_time = _getlong( &sdp->d_data[SIGN_EXPIR]);
      if( exp_time < (u_int32_t) tt.tv_sec) {
	dprintf(4, (ddt, "do_verify(): Expired signature %s %s %8d\n",
		    __p_type(type), domain, (int) (tt.tv_sec-exp_time) )) ;
	ret = NOT_OK;      
      }
    }
    if( !ret && key->axfr) {
      /* put signature in data to hash for AXFR */ 
      append_rr( domain, sdp, sig_data.ttl);
      if((ret = dnssec_verify((SIG_MODE_AXFR|SIG_MODE_UPDATE), sig_data.data,
			   sig_data.dat_len, key,NULL, 0))) {
	dprintf(33, (ddt, "do_verify:Verify Success %s %s %s %d\n",  
		     domain, signer,__p_type(type), ret)); 
      }
    }
#ifdef TESTING
    if( !ret ) 
      printf("Do_verify:Verify Success %s %s %s %d\n", domain, signer,
	     __p_type(type), ret); 
#endif /*  TESTING */
  }
  memset(sig_data.data, 0, sig_data.dat_len+8);
  sig_data.dat_len = 0;
  return ret; 
} /* do_verify */

/*  
 * get_pub_key(name, foot) Routine to look up and convert a 
 * key in DNS data base to key format for the verify routine 
 * This version does differenciate among MULTIPLE keys for same name
 */

static KEY *
get_pub_key(name, foot, sec)
char *name; 
u_int16_t foot;
int *sec;
{
    static struct databuf  *dp, *last_dp=NULL;
    struct namebuf *np;
    u_char *cp, alg, proto;
    static u_int16_t key_foot, flags;
    int len, key_len;
    static KEY *key=NULL;

/* 
 * quick and dirty mechanishm to return last key requested 
 */ 
    len= strlen(name); 
    if( name[len-1]=='.') name[--len]='\0';
    if( key && key->foot == foot && !strcmp( key->key_name, name)) {
      *sec = last_dp->d_sec;
      return key;
    }
/* 
 * Look up key 
 */ 
    dp = nt_lookup( name, T_KEY, &np);
    if( dp == NULL) {
      printf("get_pub_key: No KEY records stored under %s\n",name);
      *sec = DB_S_PENDING; /* I have no key to make determination on */
      return NULL;
    }
    *sec = DB_S_INVALID;  /* Assume KEY is bad by default */
#ifdef ROUND_ROBIN
    for( ; dp ; dp=dp->d_next)
      if( dp->d_type == T_KEY) {
#else 
    for( ; (dp != NULL) && (dp->d_type == T_KEY); dp=dp->d_next){  
#endif
      cp = dp->d_data;
      key_foot = (u_int16_t) _getshort(&dp->d_data[dp->d_size-3]);
      if( key_foot == foot ) {     /* foot print match  */
	flags = _getshort( cp);    /* new flags size is 16 bits */
	cp += sizeof(short);
	if( flags & (SIGN_F_SIGNATORY | SIGN_F_ZONE)) {
	   /* XXXX need better checks on names and scopes of KEYS NOT HERE */
	  proto =  *cp++;
	  alg   =  *cp++;
	  if( alg != 1) { 
	    printf("get_pub_key: Unknown algorithm %d\n", alg);
	    return NULL; 
	  }
	  key_len = dp->d_size - sizeof(flags) -sizeof(alg)-sizeof(proto);  
	  key     = dnssec_pubkey( (int) alg, key_len, 
				  dp->d_data+(dp->d_size-key_len), name);
	  if( key) {
	    key->flags = flags;
	    key->ttl   = dp->d_ttl;
	    *sec       = dp->d_sec;
	  }
	  else
	    *sec       = DB_S_UNKNOWN;
	  last_dp    = dp;
	  return key;
	}  /*  if(flags */
	else 
	  printf( "get_pub_key:Key %s not allowed to sign %x %x\n", name,flags,
	       SIGN_F_SIGNATORY | SIGN_F_ZONE);
      }                  /* if( key_foot ==   Foot print does not match */
    }    /* for (dp=np   */
    *sec = DB_S_PENDING ; 
    return NULL;    /*  Could not find a KEY */ 
}

/* 
 * verify_rec() Function to look up records and check if the
 *              signature supplied makes them VERIFIED 
 *              if the records can not be found           return NOT_OK 
 *              if no records correspond to the signature return NOT_OK
 *              if the dnssec_verify fails                   return NOT_OK
 *              else                                      return OK
 *
 */
int 
verify_rec( origin, domain, sdp, sec) 
char *origin, *domain; 
struct databuf *sdp;
int  sec;
{
    int domain_lab, sign_lab, i, ret;
    int16_t type;
    u_int16_t foot;
    char *cp, use_name[MAXDNAME];
    struct namebuf *np = NULL;
    struct databuf  *dp, *cdp;        /* data records */
    char vr_signer[MAXDNAME];
    KEY * key;
    u_int32_t sign_ttl;

  /* first step look up domain */
    type = _getshort(sdp->d_data);
    sdp->d_sec = DB_S_INVALID;     /* assume check is going to fail */
    if( type == T_SIG)  return NOT_OK;
    else if( type != T_AXFR) {
      dp = nt_lookup( domain, type, &np);

      if( np == NULL) {
	printf("verify_rec: NO such name %s\n", domain); 
	return NOT_OK; 
      }
      if( dp == NULL) {
	printf("verify_rec: NO records of type %s for name %s\n", 
	       __p_type(type),  domain); 
	return NOT_OK; 
      }
    }
    if( sdp->d_data[SIGN_SIGNER] == '.' || sdp->d_data[SIGN_SIGNER] == 0x0 )  
      strcpy(vr_signer,"");
    else     strcpy(vr_signer, &sdp->d_data[SIGN_SIGNER]);
    foot = (u_int16_t) _getshort(&sdp->d_data[SIGN_FOOT]);

      /*set the key of the signer*/
    if((key = get_pub_key(vr_signer, foot, &sec))==NULL) { 
      u_int32_t exp_time;
      printf("verify_rec: Did not find key for %s with footprint %d\n",
	     vr_signer, foot);
      /* now check if the signature has expired */
      exp_time = _getlong( &sdp->d_data[SIGN_EXPIR]);
      if( exp_time < (u_int32_t) tt.tv_sec) {
	dprintf(4, (ddt, "verify_rec(): Signature expired %s %s %8d\n",
		    __p_type(type), domain, (int)(tt.tv_sec-exp_time) )) ;
	verify_del( np, domain, type, sdp);
	return NOT_OK;
      }
      else {
	sdp->d_sec = sec = DB_S_PENDING;
	printf("verify_rec: %s %s marked as postponed\n", domain,
	       __p_type(type)); 
	return OK;
      }
    }
    if(  set_sig_prefix(sdp) == 0) { /* set the prefix from */
      printf("verify_rec: failed set_sig_prefix()\n");
      return NOT_OK; 				  /* the signature */
    }
    if( key_file && key->axfr == 0) key->axfr = do_axfr;

  /* Check if these records have been verified if
	   if signatures are identical               return OK
	      (I should realy add OK_NOT_ADD)             */
    sign_lab =  (int) ((u_char) sdp->d_data[SIGN_LABELS]);
    domain_lab =(int) dn_count_labels(domain);
    if( type == T_AXFR || sign_lab == domain_lab)/* required by Jan 96 change*/
      strcpy(use_name, domain);
    else if( sign_lab < domain_lab) {
      for( cp= domain, i = domain_lab-sign_lab; i>0; i--, cp++)
	cp = strchr(cp,'.');
      strcpy(use_name,"*.");
      strcat(use_name, cp);
      dprintf(4, ( ddt, "verify_rec: doing wildcard sub %d %s < %d  => %s\n",
	     domain_lab, domain, sign_lab, use_name));
    }
    else {
      printf("verify_rec: Invalid label count RR=%3d < SIG=%3d\n",
	     domain_lab, sign_lab);
      return NOT_OK;
    }
	     
  /* Marshall the records into verify queue and call dnssec_verify() 
     Extend routine to handle AXFR correctly */ 
    sign_ttl = _getlong( &sdp->d_data[SIGN_OTTL]);
    if( type != T_AXFR && type != T_NXT){
      for( ;  dp != NULL ; dp = dp->d_next)
	  if( dp->d_type == type) /* allways put all RR's in the bucket */
	    put_rr( use_name, dp, sign_ttl) ;     
    }
    else if( type == T_NXT) {
      struct databuf *dp2;
      /* find if there is a second NXT record */
      for( dp2 =dp->d_next; dp2 && dp2->d_type != T_NXT; dp2= dp2->d_next);
      if( dp2) {
	/* only use one NXT record for this SIG, which one depends on the 
	 * SIGNER's name and the value of the RDATA field 
	 * the only why to be sure is to check one and if that fails use
	 * the other one */ 
	cdp = nxt_check( use_name, domain, dp, dp2, sdp);
	if( cdp == NULL) { /* can not determine wich one, use brute force */
	  put_rr( use_name, dp, sign_ttl);
	  if( do_verify( domain, type, vr_signer, sdp, key, sign_ttl)) {
	    printf("Verify_rec: Ignore error msg above \n");
	    dp_rrcount = 0;
	    dp = dp2;
	  }
	}
	else dp = cdp;
      }
      put_rr( use_name, dp, sign_ttl);
    }
    if( (ret = do_verify( domain, type, vr_signer, sdp, key, sign_ttl)))
      dprintf(4, (ddt, "verify_rec(): Signature check failed %s %s %d\n",
	      __p_type(type), domain, ret)); 
    if( ret) {
/*   if SIG(AXFR) failes do not add record to the 
 *   database this error is most likly cause by omission or 
 *   order problem not maliciouness 
 */
      if( type != T_AXFR) {
      /* Delete all RR, for SIG(RR) that failed and are not marked VALID 
         Question what about RR's marked INSECURE ?? XXXX */
	verify_del( np, domain, type, sdp);
      }
      else {/* type == T_AXFR I think this is correct XXX */
	if( key_file) /* mark data keyfile datat that fails SIG(AXFR) as bad 
			 I can not delete as we do not have np's */
	  for(;dp_rrcount > 0; dp_rrarray[dp_rrcount]=NULL)
	    if( dp_rrarray[--dp_rrcount]->d_sec == DB_S_PENDING ) 
	      dp_rrarray[dp_rrcount]->d_sec = DB_S_INVALID;
      }
      dp_rrcount = 0; /* preventive programming */
      return VERIFY_HARD_FAIL;
    } 
#ifdef TESTING
      printf("verfify_rec():%s Success %s sec=%d\n", domain, 
	     __p_type(type), sec); 
#endif TESTING
    if( key_file && sec & DB_S_PENDING) 
      sec ^= DB_S_PENDING; /* XXXX Hack for root key */
    
/* This loop sets the security code of all the RR's that where just 
   verified dp_rrarray is set by put_rr 
 */ 
    for( sdp->d_sec = sec; dp_rrcount > 0; dp_rrarray[dp_rrcount]=NULL)
      dp_rrarray[--dp_rrcount]->d_sec = sec;
    return OK;
  } /* Verify_rec */
/* 
 * validata_rec() This routine will update all records of stored by this
 *          name and type with the flags supplied 
 * Assumption: if I call this routine I know that the parameters supplied
 *       are correct. I do not need to check with zone SOA (which I may 
 *       not have )
 * Assumption 2: If I have a key for zone that is signature key then zone is
 *       secure. 
 * Assumption 3: db_update should be able to set the flag d_sec for DB_S_SECURE
 *               or DB_S_UNSECURE
 */ 
void 
validate_rec( name, dp, flags) 
char *name; 
struct databuf *dp;
int flags;
{
  struct databuf *ndp;
  struct namebuf *np;
  u_int32_t ttl, exp;
  short type ; 
  type = dp->d_type == T_SIG ? _getshort( dp->d_data) : dp->d_type;
  ndp = nt_lookup( name, type, &np); 
  dp->d_sec = flags;
  exp = _getlong(&dp->d_data[SIGN_EXPIR]);
  if( exp < tt.tv_sec ) {  /* Delete if this data has expired */
    delete_all( np, dp->d_class, type);
    return ; 
  }                
  ttl = _getlong(&dp->d_data[SIGN_OTTL]); /* set TTL's  correctly */
  if( dp->d_zone <= DB_Z_CACHE) {        /* for remote data  set time*/
    ttl += tt.tv_sec;              
    if (exp < ttl) 
      ttl = exp;
  }
  else {              /* for local data set no of sec to live */
    if( ttl+tt.tv_sec > exp) 
      ttl = exp-tt.tv_sec;
  }
  dp->d_ttl = dp->d_ttl < ttl ? dp->d_ttl : ttl;
  for( ; ndp != NULL && ndp != dp ; ndp= ndp->d_next) {
    ndp->d_sec = flags; 
    ndp->d_ttl = ndp->d_ttl < ttl ? ndp->d_ttl : ttl;
  }
}

static void 
verify_del( np, domain, type, sdp) 
     struct namebuf *np;
     char *domain; 
     struct databuf *sdp;
     int16_t type;
{
  struct databuf *dp, *odp;
  dp = np ? np->n_data : nt_lookup(domain, 0, &np);
  for( odp = NULL;  dp ; )  
    if (dp->d_type == type || 
	(dp->d_type == T_SIG && 
	 (type == (int16_t) _getshort((char *) dp->d_data)) ) ) 
      if( dp->d_sec & (DB_S_PENDING | DB_S_UNKNOWN | DB_S_INVALID))
	dp = rm_datum( dp, np, odp);
      else {
	printf(
	"verify_del(): Attemting to delete a STATIC/VALID/INSECURE RR %s %s\n",
	       domain, __p_type(type));
	odp = dp, dp = dp->d_next;
      }
    else 
      odp = dp, dp = dp->d_next;
  return;
}


/* Function that tells us which dp record goes with this signature 
   if it can not determine then it will return the first one 
   what I need to do is to find both the NXT records and both 
   signature records (if available) and for the given signature record 
   find out which dp record goes with that signature
 */
struct databuf *
nxt_check( name, domain, dp1, dp2, sdp)
char *name, *domain;
struct databuf *dp1, *dp2, *sdp;
{
  int la,qn; 
  int n1=0, n2=0;
  int s1=0, s2=0, sx=0,nx=0;
  struct databuf *asdp=NULL;
  qn = (int) dn_count_labels(domain); 
  la = (int) dn_count_labels(name);
  if( qn < la) {
    printf( "nxt_check(): Bad query name %s %s\n", domain, name); 
    return NULL;
  }
  else { 
    n1 = dn_count_labels( dp1->d_data);
    n2 = dn_count_labels( dp2->d_data);
    s1 = dn_count_labels( &sdp->d_data[SIGN_SIGNER]);
    asdp =  dp2->d_next != sdp ? dp2->d_next : sdp->d_next; 
    if( asdp == NULL || asdp->d_type != T_SIG || 
       _getshort( asdp->d_data) != T_NXT ) { 
      printf( "nxt_check(): Could not find second SIG(NXT) %s %s %s\n", 
	     domain, name, asdp ? __p_type(asdp->d_type) : "NULL"); 
      asdp = NULL;
    }
    s2 = asdp ? dn_count_labels( &asdp->d_data[SIGN_SIGNER]): 0;
    nx = sx = 0;
    if( qn == la && s1 == s2) return NULL; /* I have to brute force this one*/
    if( qn > la) { 
      if( qn == n1)      nx = 1;
      else if( qn == n2) nx = 2;
      else {
	if(      !strcasecmp( name, dp1->d_data))       nx = 1;
	else if( !strcasecmp( name, dp2->d_data))       nx = 2;
	if( !strcasecmp( name, &sdp->d_data[SIGN_SIGNER]))       sx = 1;
	else if( !strcasecmp( name, &asdp->d_data[SIGN_SIGNER])) sx = 2;
      }
      if( sx ==0 && s1 > s2)      sx =1;
      else if( sx ==0 && s1 < s2) sx =2;
    }
    else { 
      if( n1 > n2)                             nx = 2; 
      else if( n1 < n2)                        nx = 1; 
      else if( !strcasecmp(name, dp1->d_data)) nx = 2;
      else                                     nx = 1;
      if( s1 < s2)      sx = 2;
      else if( s1 > s2) sx = 1;
    }
  }
  if( sx ==1 )
    if( nx ==2 )  return dp2;
    else  return dp1;
  if( sx ==2 ) 
    if( nx ==2 )  return dp2; /* was nx == 1 */
    else  return dp1;
  else return NULL;       /* if I do not know which one or have to do both */
}


#endif /* DNS_SECURITY */
